/*
 * rect32.c - PM VNC Viewer, Pixel Handling for 32 bit true color
 */

#include <stdio.h>
#include <stdlib.h>

#define INCL_PM
#include <os2.h>

#include "pmvncdef.h"

/*
 * rbuff is used to parse CoRRE, it requires buffer of
 * 255 * 255 * 32 bits (nearly 64KB * 4).
 */
 
#define RBUFSIZ     (1024 * 256)

static  UCHAR   rbuff[RBUFSIZ] ;        /* recv buffer  */

/*
 * buffer for bitmap scan conversion
 */
 
#define MAXSCAN     (4096 * 8)

static  UCHAR   ibuff[MAXSCAN] ;        /* scan buffer  */

#define MINBUF  MAXSCAN     /* min(MAXSCAN, RBUFSIZ)    */

/*
 * height of RFB is used for convert Y coordinate
 */

static  int     rfbHeight = 0 ;

static  int     queryHeight(void)
{
    SIZEL   siz ;
    protoQueryDesktop(&siz) ;
    return siz.cy ;
}

#define RFBHEIGHT ((rfbHeight > 0) ? rfbHeight : (rfbHeight = queryHeight()))

/*
 * rectRaw - raw rectangle encoding
 *      read RawRect data from Server and convert to PM Bitmap
 */

static  void    rawConv(rfbRectangle *r, int lines, PUCHAR buff)
{
    PUCHAR  sp, dp ;
    int     i, j   ;
    int     bytesPerLine ;
    
    /*
     * convert to PM bitmap, reverse vertical order and align to ULONG
     */
     
    bytesPerLine = ((r->w * 3) + 3) & ~0x03 ;
    
    for (i = 0, sp = buff ; i < lines ; i++) {
	dp = &ibuff[bytesPerLine * (lines - i - 1)] ;
	for (j = 0 ; j < r->w ; j++) {
	    dp[0] = sp[0] ;     /* Blue     */
	    dp[1] = sp[1] ;     /* Green    */
	    dp[2] = sp[2] ;     /* Red      */
	    sp += 4 ;
	    dp += 3 ;
        }
    }
    bmpRectDraw(r->x, (RFBHEIGHT - r->y - lines), r->w, lines, ibuff) ;
    return ;
}
 
static  BOOL    rectRaw(rfbRectangle *r)
{
    int     bytesPerLine, linesToRead ;
    
    bytesPerLine  = r->w * 4 ;      /* for 32 bits depth specific   */
    linesToRead   = MINBUF / bytesPerLine ;
    
    if (linesToRead == 0) {
        netFail("rectRaw - scanline too large, recompile!!") ;
	return FALSE ;
    }

    while (r->h > 0) {
        if (linesToRead > r->h) {
            linesToRead = r->h ;
        }
        if (netRecv(rbuff, bytesPerLine * linesToRead) != TRUE) {
	    netFail("failed to recv Raw Scanline") ;
	    return FALSE ;
	}
        rawConv(r, linesToRead, rbuff) ;
	r->h -= linesToRead ;
	r->y += linesToRead ;
    }
    return TRUE ;
}

/*
 * rectCopy - copy rect encoding
 */
 
static  BOOL    rectCopy(rfbRectangle *r)
{
    rfbCopyRect cr  ;
    
    if (netRecv((PUCHAR) &cr, sz_rfbCopyRect) != TRUE) {
        netFail("failed to recv CopyRect") ;
	return FALSE ;
    }
    cr.srcX = swap16(cr.srcX) ;
    cr.srcY = swap16(cr.srcY) ;
    
    bmpRectCopy(r->x, (RFBHEIGHT - r->y - r->h), r->w, r->h,
                        cr.srcX, (RFBHEIGHT - cr.srcY - r->h)) ;
    return TRUE ;
}

/*
 * rectRRE - RRE encoding
 */

static  BOOL    rectRRE(rfbRectangle *r)
{
    rfbRREHeader    hdr ;
    rfbRectangle    subrect ;
    CARD32          color   ;
    int             i       ;
    
    if (netRecv((PUCHAR) &hdr, sz_rfbRREHeader) != TRUE) {
        netFail("failed to recv RREHeader") ;
	return FALSE ;
    }
    hdr.nSubrects = swap32(hdr.nSubrects) ;

    if (netRecv((PUCHAR) &color, sizeof(CARD32)) != TRUE) {
        netFail("failed to recv BG color") ;
	return FALSE ;
    }

    bmpRectColor(color & 0x00ffffff) ;
    bmpRectFill(r->x, (RFBHEIGHT - r->y - r->h), r->w, r->h) ;
    
    for (i = 0 ; i < hdr.nSubrects ; i++) {
        if (netRecv((PUCHAR) &color, sizeof(CARD32)) != TRUE) {
	    netFail("failed to recv sub color") ;
	    return FALSE ;
	}
	if (netRecv((PUCHAR) &subrect, sz_rfbRectangle) != TRUE) {
	    netFail("failed to recv subrect") ;
	    return FALSE ;
	}
	subrect.x = r->x + swap16(subrect.x) ;
	subrect.y = r->y + swap16(subrect.y) ;
	subrect.w = swap16(subrect.w) ;
	subrect.h = swap16(subrect.h) ;

        bmpRectColor(color & 0x00ffffff) ;
        bmpRectFill(subrect.x, 
	    (RFBHEIGHT - subrect.y - subrect.h), subrect.w, subrect.h) ;
    }
    return TRUE ;
}

/*
 * rectCoRRE - CoRRE encoding
 */
 
static  BOOL    rectCoRRE(rfbRectangle *r)
{
    rfbRREHeader    hdr ;
    rfbRectangle    subrect ;
    PUCHAR          sp      ;
    CARD32          color   ;
    int             i       ;
    
    if (netRecv((PUCHAR) &hdr, sz_rfbRREHeader) != TRUE) {
        netFail("failed to recv RREHeader") ;
	return FALSE ;
    }
    hdr.nSubrects = swap32(hdr.nSubrects) ;

    if (netRecv((PUCHAR) &color, sizeof(CARD32)) != TRUE) {
        netFail("failed to recv BG color") ;
	return FALSE ;
    }

    bmpRectColor(color & 0x00ffffff) ;
    bmpRectFill(r->x, (RFBHEIGHT - r->y - r->h), r->w, r->h) ;
    
    if (netRecv(rbuff, hdr.nSubrects * 8) != TRUE) {
        netFail("failed to recv subrects") ;
	return FALSE ;
    }

    for (i = 0, sp = rbuff ; i < hdr.nSubrects ; i++, sp += 8) {
        color = *(CARD32 *) sp ;
	subrect.x = r->x + sp[4] ;
	subrect.y = r->y + sp[5] ;
	subrect.w = sp[6] ;
	subrect.h = sp[7] ;
        bmpRectColor(color & 0x00ffffff) ;
        bmpRectFill(subrect.x, 
	    (RFBHEIGHT - subrect.y - subrect.h), subrect.w, subrect.h) ;
    }
    return TRUE ;
}

/*
 * rectHex - Hextile encoding
 */

static  CARD32  tileFg = 0xffffffff ;
static  CARD32  tileBg = 0x00000000 ;

static  BOOL    hextileDraw(int x, int y, int w, int h)
{
    PUCHAR  sp, dp ;
    int     i, j   ;
    int     bytesPerLine ;

    bytesPerLine = ((w * 3) + 3) & ~0x03 ;

    if (netRecv(rbuff, w * h * 4) != TRUE) {
        netFail("failed to hextile raw") ;
	return FALSE ;
    }
    
    for (i = 0, sp = rbuff ; i < h ; i++) {
        dp = &ibuff[bytesPerLine * (h - i - 1)] ;
	for (j = 0 ; j < w ; j++) {
	    dp[0] = sp[0] ;
	    dp[1] = sp[1] ;
	    dp[2] = sp[2] ;
	    sp += 4 ;
	    dp += 3 ;
	}
    }
    bmpRectDraw(x, (RFBHEIGHT - y - h), w, h, ibuff) ;
    return TRUE ;
}

static  BOOL    hextileFill(int x, int y, int w, int h)
{
    bmpRectColor(tileBg & 0x00ffffff) ;
    bmpRectFill(x, (RFBHEIGHT - y - h), w, h) ;
    return TRUE ;
}

static  BOOL    hextileSubSolid(int x, int y, int w, int h)
{
    CARD8   nSubrects ;
    PUCHAR  sp        ;
    int     sn, sx, sy, sw, sh ;

    bmpRectColor(tileBg & 0x00ffffff) ;
    bmpRectFill(x, (RFBHEIGHT - y - h), w, h) ;

    /*
     * fill subrects with foregournd color
     */

    if (netRecv((PUCHAR) &nSubrects, 1) != TRUE) {
        netFail("failed to recv hextile nSubrects") ;
	return FALSE ;
    }
    if (netRecv((PUCHAR) rbuff, nSubrects * 2) != TRUE) {
        netFail("failed to recv hextile Subrects") ;
	return FALSE ;
    }

    bmpRectColor(tileFg & 0x00ffffff) ;

    for (sn = 0, sp = rbuff ; sn < nSubrects ; sn++, sp += 2) {
        sx = rfbHextileExtractX(sp[0]) ;
        sy = rfbHextileExtractY(sp[0]) ;
	sw = rfbHextileExtractW(sp[1]) ;
	sh = rfbHextileExtractH(sp[1]) ;
        bmpRectFill((x + sx), (RFBHEIGHT - y - sy - sh), sw, sh) ;
    }    
    return TRUE ;
}

static  BOOL    hextileSubColor(int x, int y, int w, int h)
{
    CARD8   nSubrects ;
    PUCHAR  sp        ;
    PULONG  cp        ;
    int     sn, sx, sy, sw, sh ;
        
    bmpRectColor(tileBg & 0x00ffffff) ;
    bmpRectFill(x, (RFBHEIGHT - y - h), w, h) ;

    /*
     * fill subrects with specified color
     */

    if (netRecv((PUCHAR) &nSubrects, 1) != TRUE) {
        netFail("failed to recv hextile nSubrects") ;
	return FALSE ;
    }
    if (netRecv((PUCHAR) rbuff, nSubrects * 6) != TRUE) {
        netFail("failed to recv hextile Subrects") ;
	return FALSE ;
    }

    for (sn = 0, sp = rbuff ; sn < nSubrects ; sn++, sp += 6) {
        cp = (PULONG) sp ;
        sx = rfbHextileExtractX(sp[4]) ;
        sy = rfbHextileExtractY(sp[4]) ;
	sw = rfbHextileExtractW(sp[5]) ;
	sh = rfbHextileExtractH(sp[5]) ;

	bmpRectColor(*cp & 0x00ffffff) ;
	bmpRectFill((x + sx), (RFBHEIGHT - y - sy - sh), sw, sh) ;
    }    
    return TRUE ;
}

static  BOOL    rectTile(rfbRectangle *r)
{
    int     x, y, w, h ;
    CARD8   subencoding ;
    
    for (y = r->y ; y < (r->y + r->h) ; y += 16) {
        for (x = r->x ; x < (r->x + r->w) ; x += 16) {
            w = h = 16 ;
	    if ((r->x + r->w - x) < 16) {
	        w = r->x + r->w - x ;
	    }
	    if ((r->y + r->h - y) < 16) {
	        h = r->y + r->h - y ;
	    }
	    if (netRecv((PUCHAR) &subencoding, 1) != TRUE) {
	        netFail("failed to recv sub encoding") ;
		return FALSE ;
	    }
	    if (subencoding & rfbHextileRaw) {
	        if (hextileDraw(x, y, w, h) != TRUE) {
		    return FALSE ;
		}
		continue ;
	    }
	    if (subencoding & rfbHextileBackgroundSpecified) {
	        if (netRecv((PUCHAR) &tileBg, sizeof(CARD32)) != TRUE) {
		    netFail("failed to recv BG color") ;
		    return FALSE ;
		}
	    }
	    if (subencoding & rfbHextileForegroundSpecified) {
	        if (netRecv((PUCHAR) &tileFg, sizeof(CARD32)) != TRUE) {
		    netFail("failed to recv FG color") ;
		    return FALSE ;
		}
	    }
	    if ((subencoding & rfbHextileAnySubrects) == 0) {
	        if (hextileFill(x, y, w, h) != TRUE) {
		    return FALSE ;
		}
            } else if (subencoding & rfbHextileSubrectsColoured) {
	        if (hextileSubColor(x, y, w, h) != TRUE) {
		    return FALSE ;
		}
	    } else {
	        if (hextileSubSolid(x, y, w, h) != TRUE) {
		    return FALSE ;
		}
	    }
        }
    }
    return TRUE ;
}

/*
 * ctxPix32 - Drawing Context for 32 bit true color
 */

VNCCTX      ctxPix32 = {
    /* bitsPerPixel */  32,         /* Current pixel format will fit    */
    /* depth        */  32,         /* to PM's RGB structure with no    */
    /* bigEndian    */  0,          /* conversions.                     */
    /* trueColour   */  1,          /* It reduces client side load.     */
    /* redMax       */  0x00ff,     /* But for reduce network traffic,  */
    /* greenMax     */  0x00ff,     /* 8 bits BGR233 will be nice.      */
    /* blureMax     */  0x00ff,
    /* redShift     */  16,
    /* greenShift   */  8,
    /* blueShift    */  0,
    /* pad1, pad2   */  0, 0,
    /* rectRaw      */  rectRaw,
    /* rectCopy     */  rectCopy,
    /* rectRRE      */  rectRRE,
    /* rectCoRRE    */  rectCoRRE,
    /* rectTile     */  rectTile
} ;
